# ZPT4Forms
# Forms class
# (c) 2002 n@work Internet Informationssysteme GmbH
#          Andreas Elvers
#          andreas@work.de

from Persistence import Persistent
from AccessControl import ClassSecurityInfo
from Globals import InitializeClass
from TALDefs import FORMSError
from types import StringType
import re
from Products.Forms4ZPT.tools.Loggerr import loggerr

reSplitCheck      = re.compile("([^\' ]+|\'[^\']+\')(?:\s+|$)")
reSplitCheckOrder = re.compile("(\d+)$")
reMatchEmail      = re.compile("^[\w.-]+\@(\w+[\w-]*\.)+\w+$")
reMatchFloat      = re.compile("^[+\-]?\d+(?:\.\d+)?$")
reMatchInt        = re.compile("^[+\-]?\d+$")

# remove sourrounding quotes 
def _remQuotes(string):
    if string[0] == "'" and string[-1] == "'":
        return string[1:-1]
    else:
        return string

class Forms(Persistent):
    "This class acts as a container for all forms checking stuff"
    _options=None
    _fields=None
    security = ClassSecurityInfo()
    
    def __init__(self,template):
        self._forms = {}
        self._template = template

    def __repr__(self):
        # outdated FIXME
        return "options:\n%s\nchecks:\n%s" % (str(self._options),
                                              str(self._checks))

    def _setOption(self,formname,option,**kw):
        forms = self._forms
        if not forms.has_key(formname):
            self.initForm(formname)
        forms[formname]['options'][option] = {}
        option = forms[formname]['options'][option]
        for key in kw.keys():
            option[key] = kw[key]
        self._p_changed = 1
            
    def _setCheck(self,form,field,check,**kw):
        if not self._forms[form]['fields'].has_key(field):
            self.initField(form,'',field)
        checks = self._forms[form]['fields'][field]['checks']
        dict = {'name':check}
        dict.update(kw)
        checks.append(dict)
        # sort checks
        checks.sort(lambda x,y: x['order'] - y['order'])
        self._p_changed = 1

    #def processRequest(self,request):
    #    formdata = request.form
    #    formname = formdata['f4zpt_formname']
    #    fields = self._forms[formname]['fields']
    #
        #for field in fields.keys():
        #    if not formdata.has_key(field) and fields.has_key(field):
        #        try:
        #            formdata.set(field,fields[field]['default'][:])
        #        except:
        #            raise str(fields[field]['default'][:])
                

    security.declarePublic('getRequestFieldValue')
    def getRequestFieldValue(self,formName,fieldName,request,override=None):
        "returns the posted value of a form field"
        if request.environ['REQUEST_METHOD'] == 'POST' and not override:
            try:
                return request.form[fieldName]
            except KeyError:
                return ""
        else:
            if request.has_key('f4zpt_defaults'):
                # since this is a GET or override try to return
                # runtime default if available
                try:
                    return  request['f4zpt_defaults'][fieldName][0]
                except KeyError:
                    pass

            try:
                return self._forms[formName]['fields'][fieldName]['default'][0]
            except IndexError:
                return ""


    security.declarePublic('getFieldLabel')
    def getFieldLabel(self,formName,fieldName,value):
                
        try:
            return self._forms[formName]['fields'][fieldName]['label'][value]
        except KeyError:
            return value
        
    security.declarePublic('getCheckedAttr')
    def getCheckedAttr(self, formName, fieldName, thisValue, request, override=None):
        "checks wether the 'checked' attribute should be set for this field"
        reqForm = request.form

        if request.environ['REQUEST_METHOD'] == 'POST' and not override:
            try:
                field = request.form[fieldName]
            except KeyError:
                return 0
        elif request.has_key('f4zpt_defaults'):
            try:
                field = request['f4zpt_defaults'][fieldName]
            except KeyError:
                return 0

        else:
            field = self._forms[formName]['fields'][fieldName]['default']

        try:
            if thisValue in field:
                return 1
            else:
                return 0
        except TypeError:
            # seems to be a simple type
            if thisValue == field:
                return 1
            else:
                return 0

            
        
    def initForm(self,formName):
        form = self._forms[formName] = {}
        form['options'] = {}
        form['checks'] = []
        form['fields'] = {}
        self._p_changed = 1

    security.declarePublic('setRuntimeDefaults')
    def setRuntimeDefaults(self,form,field,defaults,request):
        # runtime defaults are needed if forms:default is used
        if not request.has_key('f4zpt_defaults'):
            request.set('f4zpt_defaults',{})
        reqDefaults = request.get('f4zpt_defaults')

        if type(defaults) == StringType:
            reqDefaults[field] = [defaults[:],]
        else:
            reqDefaults[field] = defaults[:]
        

    def initField(self,formname,fieldtype,field,valueLabel={},default=""):
        """called via TALGenerator
           sets default values of form-fields. These
           are needed for the first time viewing the
           form (GET) (if no runtime default is set via forms:default).
           All later viewing will retrieve
           values from REQUEST.
           Normally initField is called with a valueLabel dict.
           When called with empty valueLabels but with default (simple type
           or list) these values are copied to valueLabel.
           """

        try:
            field = self._forms[formname]['fields'][field]
            if not field['fieldtype']:
                # this may happen due to a late initField
                # in TALGenerator for select and textarea
                # with checks
                field['fieldtype'] = fieldtype
        except KeyError:
            self._forms[formname]['fields'][field] = field = {}
            field['label'] = {}
            field['default'] = []
            field['checks'] = []
            field['fieldtype'] = fieldtype
            
        # add value/s and possible label
        if valueLabel:
            field['label'].update(valueLabel)

        if default:

            if not valueLabel:

                if type(default) == StringType:
                    field['label'][default]  = ''
                else:
                    for value in default:
                        field['label'][value] = ''
            
            if fieldtype in ('checkbox', 'option', 'textarea','select'):
                # adding defaults for possibly multiple values
                if type(default) == StringType:
                    field['default'].append(default[:])
                else:
                    field['default'].extend(default[:])

            else:
                # adding default for single default value
                field['default'] = default[:]
        self._p_changed = 1

    security.declarePublic('getFields')
    def getFields(self):
        "returns a list of form/field/type dict"
        fields=[]
        for form in self._forms.keys():
            for field in self._forms[form]['fields'].keys():
                fdata = self._forms[form]['fields'][field]
                if len(fdata['label']):
                    for value in fdata['label'].keys():
                        fields.append({'form':form,
                                       'field':field,
                                       'type':fdata['fieldtype'],
                                       'value':value,
                                       'label':fdata['label'][value],
                                       'default':fdata['default'],
                                       'checks':fdata['checks']})
                else:
                    fields.append({'form':form,
                                   'field':field,
                                   'type':fdata['fieldtype'],
                                   'value':'',
                                   'label':'',
                                   'default':fdata['default'],
                                   'checks':fdata['checks']
                                   })
        return fields
        
    # options

    def setOptPersistent(self,form,ttl,submits):
        """set a form persistent for ttl seconds. A ttl of 0 is
           considered an infinite time to live."""
        self._setOption(form,'persistent',ttl=ttl,submits=submits)

    def setOptExternalCheck(self, form, methodName):
        self._setOption(form,'extcheck', methodName = methodName)

    def setOptExit(self,form,method):
        self._setOption(form,'exitmethod',method=method)

    def getOption(self, formName, optionName):
        return self._forms[formName]['options'][optionName]

    security.declarePublic('getOptions')
    def getOptions(self):
        "returns a list of form/option/value dict"
        options = []

        for form in self._forms.keys():
            for option, value in self._forms[form]['options'].items():
                options.append({'form':form,
                                'option':option,
                                'value':str(value)})
                
        return options
                   

    def applyFormChecks(self, context, request):
        """apply all defined checks in correct order"""
        form = request.form
        formToCheck = form['f4zpt_formname']
        errors = []
        
        for fieldName in self._forms[formToCheck]['fields'].keys():
            field = self._forms[formToCheck]['fields'][fieldName]
            checks = field['checks']
            for check in checks:
                message = ""
                fieldVal = form[fieldName]
                invert = check['invert']
                # check empty
                if check['name'] == 'empty':
                    try:
                        if not fieldVal:
                            message = check['message']
                    except KeyError:
                        message = check['message']
                #check email
                elif check['name'] == 'email':
                    try:
                        if not reMatchEmail.match(fieldVal):
                            message = check['message']
                    except KeyError:
                        message = check['message']
                # check re
                elif check['name'] == 're':
                    try:
                        if not check['compiledRegExp'].search(fieldVal):
                            message = check['message']
                    except KeyError:
                        message = check['message']
                # check range
                elif check['name'] == 'range':
                    try:
                        if check['numberFrom'] > float(fieldVal) or float(fieldVal) > check['numberTo']:
                            message = check['message']
                    except ValueError:
                        message = check['message']
                # check >
                elif check['name'] == 'gt':
                    try:
                        if not float(fieldVal) > float(check['checkval']):
                            message = check['message']
                    except ValueError:
                        message = check['message']
                # check <
                elif check['name'] == 'lt':
                    try:
                        if not float(fieldVal) < float(check['checkval']):
                            message = check['message']
                    except ValueError:
                        message = check['message']
                # check ==
                elif check['name'] == 'eq':
                    try:
                        if fieldVal != check['checkval']:
                            message = check['message']
                    except ValueError:
                        message = check['message']
                elif check['name'] == 'type':
                    if check['checkval'] == 'float':
                        regexp = reMatchFloat
                    if check['checkval'] == 'int':
                        regexp = reMatchInt
                    if not regexp.search(fieldVal):
                        message = check['message']
                else:
                    raise FORMSError("unkown check '%s' received (internal error)" % check['name'])

                # do NOT
                if invert:
                    if message:
                        message = ""
                    elif not message:
                        message = check['message']

                if message:
                    errors.append({'field':fieldName,
                                   'checktype':check['name'],
                                   'message':message})
        if not errors and self.hasExternalCheck(formToCheck):
            # FIXME: right now, the external check will only be
            # called if all internal checks have been passed.
            errors = self.executeExternalCheck(formToCheck, context, request)
            
        return {formToCheck:{'list':errors}}
                                       

    def executeExternalCheck(self, formToCheck, context, request):
        methodName = self.getOption(formToCheck, 'extcheck')['methodName']
        method = getattr(context,methodName)
        errorTuples = method()
        errors = []
        for field,message in errorTuples:
            errors.append({'field':field,
                           'message':message,
                           'checktype':'external'})
        return errors

    def hasExternalCheck(self, formToCheck):
        if self._forms[formToCheck]['options'].has_key('extcheck'):
            return 1

    def hasExit(self, formToCheck):
        if self._forms[formToCheck]['options'].has_key('exitmethod'):
            return 1


    def doExit(self, formName, context, request):
        methodName = self.getOption(formName, 'exitmethod')['method']
        method = getattr(context,methodName)
        return method(context,request)

    # set checks in forms structure for individual form field
    # for processing see applyFormChecks()
    
    def setCheck(self, form, field, checkattr, checkval, position):
        args = map(lambda x:_remQuotes(x),
                   reSplitCheck.findall(checkval))
        if args[0] == 'not':
            invert = 1
            del args[0]
        else:
            invert = 0
            
        checkMethod = args[0]
        checkOrder = int(reSplitCheckOrder.findall(checkattr)[0])
        if checkMethod == 'empty':
            self.setChkEmpty(form,field,checkOrder,invert,position,args)
        elif checkMethod == 'email':
            self.setChkEmail(form,field,checkOrder,invert,position,args)
        elif checkMethod == 'range':
            self.setChkRange(form,field,checkOrder,invert,position,args)
        elif checkMethod == 're':
            self.setChkRe(form,field,checkOrder,invert,position,args)
        elif checkMethod == '>':
            self.setChkGt(form,field,checkOrder,invert,position,args)
        elif checkMethod == '<':
            self.setChkLt(form,field,checkOrder,invert,position,args)
        elif checkMethod == '==':
            self.setChkEq(form,field,checkOrder,invert,position,args)
        elif checkMethod == 'type':
            self.setChkType(form,field,checkOrder,invert,position,args)
        else:
            raise FORMSError('unknown check method \'%s\'' % checkMethod,
                             position)
        
    def setChkEmpty(self,form,field,order,invert,position,args):
        """check if a field is empty.
           parameter: empty errormessage"""
        if len(args) != 2:
            raise FORMSError("check empty expects 2 arguments",
                             position)
        self._setCheck(form,field,'empty',
                       message=args[1],
                       invert=invert,
                       order=order)

    def setChkEmail(self,form,field,order,invert,position,args):
        """check if a field contains an email
           parameter: email errormessage"""
        if len(args) != 2:
            raise FORMSError("check email expects 2 arguments",
                             position)
        self._setCheck(form,field,'email',
                       message=args[1],
                       invert=invert,
                       order=order)

    def setChkRange(self,form,field,order,invert,position,args):
        """check if a number value is between the two given values
           parameter: range number_from number_to error_message"""
        
        if len(args) != 4:
            raise FORMSError("check range expects 4 arguments",
                             position)
        self._setCheck(form,field,'range',
                       numberFrom = float(args[1]),
                       numberTo = float(args[2]),
                       message = args[3],
                       invert = invert,
                       order=order)

    def setChkRe(self,form,field,order,invert,position,args):
        """check if a regular expression matches a field entry"""

        if len(args) != 3:
            raise FORMSError("check re expects 3 arguments",
                             position)
        self._setCheck(form,field,'re',
                       compiledRegExp = re.compile(args[1]),
                       message = args[2],
                       invert = invert,
                       order = order)
    def setChkGt(self,form,field,order,invert,position,args):
        """check for numerical >"""
        if len(args) != 3:
            raise FORMSError("check > expects 3 arguments",
                             position)
        self._setCheck(form,field,'gt',
                       checkval = args[1],
                       message = args[2],
                       invert = invert,
                       order = order)

    def setChkLt(self,form,field,order,invert,position,args):
        """check for numerical >"""
        if len(args) != 3:
            raise FORMSError("check < expects 3 arguments",
                             position)
        self._setCheck(form,field,'lt',
                       checkval = args[1],
                       message = args[2],
                       invert = invert,
                       order = order)

    def setChkEq(self,form,field,order,invert,position,args):
        """check string equal"""
        if len(args) != 3:
            raise FORMSError("check == expects 3 arguments",
                             position)
        self._setCheck(form,field,'eq',
                       checkval = args[1],
                       message = args[2],
                       invert = invert,
                       order = order)

    def setChkType(self,form,field,order,invert,position,args):
        """check type (numerical right now)"""
        if len(args) != 3:
            raise FORMSError("check == expects 3 arguments",
                             position)
        if args[1] not in ('int','float'):
            raise FORMSError("check type only supports int and float",
                             position)
        
        self._setCheck(form,field,'type',
                       checkval = args[1],
                       message = args[2],
                       invert = invert,
                       order = order)


InitializeClass(Forms)
